#!/usr/bin/perl
# AGRIP AudioQuake Game Launcher
# Copyright 2004-2005 Matthew Tylee Atkinson
# 
# This program is part of AGRIP AudioQuake.
# 
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
# 
# See file, 'COPYING', for details.
 
use strict;
use warnings;
use Config::Tiny;


##############################
# Global Constants & Variables
##############################

# FIXME move as many of these as possible away from being global

my $config;
my $os;
my $base_path;
my $path_sep;
my $quake_prog;
my $agripdir_path;
my $autoexec_path;
my $config_path;
my( $pak0, $pak1r, $pak1m, $pak2m );
my $registered = 0;
my $mindgrid = 0;

# Constants
use constant WIN32_ENGINE_OPTS_PRE => 'cmd /c start /b /belownormal ';
use constant WIN32_ENGINE_OPTS_GL => ' -nomouse -width 640 -height 480 +set cl_confirmquit 0';
use constant WIN32_ENGINE_OPTS_VN => ' -nomouse -startwindowed +set cl_confirmquit 0';
use constant NIX_ENGINE_OPTS_STD => ' -nomouse -width 640 -height 480 +set cl_confirmquit 0';
use constant RENDER_SAFE => '1';
use constant RENDER_GL => '0';
use constant LAUNCH_GAMEDIR => '1';  # tells main_mainloop() that we're specifying a gamedir.
use constant LAUNCH_RAW => '0';  # tells main_mainloop() that we're specifying a list of raw params.

# Installer constants and variables...
use constant AS_QARC_PT1 => 'Q101_INT.1';
use constant AS_QARC_PT2 => 'Q101_INT.2';
use constant AS_QARC_FULLNAME => 'resource.1';
my $AS_Q106 = 0;


##############################
# Subroutines
##############################

sub init();						# Works out which OS we're on and sets up TTS.
								# If we're on win32; checks if it needs to install
								# the full game's data from CD.

sub init_quakeprog();		   # Works out which type of engine (software/gl) we're
								# using and calls u_update_quakeprog() accordingly.

sub datacheck();				# Works out if we need to install extra data; if so
								# it calls other functions to do this or warn the
								# user that they have an invalid setup.

sub datacheck_detectpaks($);	# Routine that detects which pakfiles we have

sub player_account();			# Ensure the user is aware that they need to set up
								# a user account on http://stats.agrip.org.uk/ .

sub player_account_update();	# Core user/pass update code.

# Main menu functions...
sub main_menu();				# The Launcher's main menu.
sub main_mainloop($$);			# The main in-game message filter.
sub main_keybindings();			# Set and review key bindings.
sub main_keybindings_edit($$);	# Edit an individual binding.
sub main_keybindings_apply($$);	# Apply a new binding (save down config files).
sub main_accessopts();			# Set and review access options.
sub main_accessopts_edit($$$);	# Edit an individual option.
sub main_accessopts_apply($$);	# Apply a new binding (save down autoexec file).
sub main_renderopts();			# Choose video/audio output settings.

# Registered game data installer...
sub install_cdgamedata();		# For non-registered users on win32.
sub install_getdrive();		 # Get a sane CD drive letter from the user.
sub install_examinedisc($);	 # Tests files on CD to determine how to install.

# QMOD installer...
sub qmod_main($);				# Main QMOD installation logic.

sub qmod_shreddir($);			# Cross-platform routine to remove a directory
								# that may contain files recursively.

sub qmod_lister();				# Lists all (probably) valid MOD gamedirs

sub qmod_cclist_config();		# Lists all mods wishing to track the default AQ
								# config.cfg file.

sub qmod_ccquery_config($);		# Returns value for a single mod instead of list
sub qmod_cclist_autoexec();		# As above
sub qmod_ccquery_autoexec($);	# As above

# Utility functions...
sub u_msg($$);					# Prints out a u_msg to STDOUT and SPEECH.

sub u_die($);					# Prints out a u_msg to STDOUT and SPEECH,
								# waits for ENTER and then exits.

sub u_getanswer(@);				# Gets a valid choice from many possible
								# answers.

sub u_filecopy($$);				# Portable way to copy text files.

sub u_update_config();			# Update mods that are watching config.cfg.
sub u_update_autoexec();		# As above for autoexec.cfg

sub u_update_quakeprog($);		# Update the $quake_prog varilable -- used
								# when changing rendering options.

sub u_get_user_pass();			# Turns the username and pwd for MAUTH into
								# a string suitable for insertion into the
								# command line.

sub u_authcheck($$);            # Check for correct user and pass.


##############################
# Order of execution
##############################

# This program acts as both the launcher and QMOD installer.
# When it starts, it checks to see if it has been passed the name of a QMOD to
# install.  On 'doze, it needs to be passed the directory of the AQ install so
# it can find the Perl libraries.
#
# start.pl can be called in different ways, as follows.
#   start.pl
#	   This goes into the main menu.
#   [nix] start.pl install qmodfile
#   [win32] start.pl install aq_basedir qmodfile
#	   These are for installing QMODs.
#   start.pl rawlaunch
#	   This is for launching ZQ with custom cmdline paramaters. It is used by
#	   test scripts (such as for testing changs to the engine, or new maps).

# Decide where to init()...
# NB: We have to go through this bit separately as we will need to tell the
#	 user using speech if anything goes wrong later on.
if( $ARGV[0] )
{
	if( $ARGV[0] eq 'install' )
	{
		if( $ARGV[1] and -d $ARGV[1] )
		{
			# 'doze filetype association calls start.pl with AQ's installation dir
			# so that we can change to it and init() (thus finding the perl libs).
			chdir $ARGV[1];
		}
	}
}
init();

# Now we've init'd, we can decide what mode to operate in...
# NB: Now we can speak to the user if there are errors.
if( $ARGV[0] )
{
	if( $ARGV[0] eq 'install' )
	{
		if( $ARGV[1] and -f $ARGV[1] )
		{
			qmod_main($ARGV[1]);
		}
		elsif( ($ARGV[1] and $ARGV[2]) and (-d $ARGV[1] and -f $ARGV[2]) )
		{
			qmod_main($ARGV[2]);
		}
		else
		{
			if( $os eq 'win32' )
			{
				u_die "The launcher was asked to start up in QMOD installation mode,\nbut the AudioQuake base directory and a QMOD file were not specified.\n";
			}
			else
			{
				u_die "The launcher was asked to start up in QMOD installation mode,\nbut a readable QMOD file was not specified.\n";
			}
		}
	}
	elsif( $ARGV[0] eq 'rawlaunch' )
	{
		my @launch_args = @ARGV[1 .. $#ARGV];
		my $launch_args = join(' ', @launch_args);
		main_mainloop(LAUNCH_RAW,$launch_args);
	}
	else
	{
		# Unknown mode specified...
		u_die "The launcher was told to start up in a mode it didn't understand -- ``" . $ARGV[0] . "''.\nDo not specify any command-line paramaters if you wish to go to the main menu.\nOther valid startup modes are as follows.\ninstall, used for installing a QMOD file.\nrawlaunch, used to start ZQuake with some custom command-line paramaters.\n", 1;
	}
}
else
{
	# No paramaters; start normally...
	datacheck();
	player_account();
	main_menu();
}


##############################
# init
##############################

sub init()
{
	if( $ENV{'COMSPEC'} and $ENV{'COMSPEC'} =~ /command|cmd/i )
	{
		# Running on win32
		$os = 'win32';
		$base_path = '.\\';
		$path_sep = '\\';

		# Read the configuration file...
		$config = Config::Tiny->read($base_path . 'start.ini') or die "The AudioQuake Launcher cannot start.\nThe settings file \"" . $base_path . "start.ini\" can't be found.\n";
		
		# Do we need to do any initialisation commands?
		system($config->{'win32'}->{'win32_init_cmd'}) if $config->{'win32'}->{'win32_init_cmd'};

		# Open the pipe to our speaker program...
		open( SPEECH, $config->{'win32'}->{'win32_talker'} ) or die "ERROR: Couldn't open \"". $config->{'win32'}->{'win32_talker'} . "\" -- please ensure you've got it set up and/or installed.\n";
	
		# Make it the default file...
		select SPEECH;
	
		# Automatically flush the buffer to it on every print...
		$| = 1;

		# Which Quake program are we using?
		init_quakeprog();

		# Set paths to key files...
		$agripdir_path = $base_path . 'id1';
		$autoexec_path = $agripdir_path . '\autoexec.cfg';
		$config_path = $agripdir_path . '\config.cfg';

		# Go to base path...
		chdir($base_path);  # FIXME necessary?

		# Move pak2.pak if need be...
		if( -e 'mindgrid-audio_quake_2003.09.22/pak2.pak' )
		{
			rename('mindgrid-audio_quake_2003.09.22/pak2.pak', 'id1/pak2.pak');
		}

		# Do we need to send an init string?
		foreach my $key ( keys %{$config->{'win32_init_strings'}} )
		{
			print SPEECH $config->{'win32_init_strings'}->{$key} . "\n";
		}
	}
	else
	{
		# Running on nix
		$os = "nix";
		$base_path = $ENV{'HOME'} . '/.zquake/';
		$path_sep = '/';

		# Read the configuration file...
		$config = Config::Tiny->read($base_path . 'start.ini') or die "The AudioQuake Launcher cannot start.\nThe settings file \"" . $base_path . "start.ini\" can't be found.\n";

		# Do we need to do any initialisation commands?
		system($config->{'nix'}->{'nix_init_cmd'}) if $config->{'nix'}->{'nix_init_cmd'};

		# Open the pipe to our speaker program...
		open( SPEECH, $config->{'nix'}->{'nix_talker'} ) or die "ERROR: Couldn't open \"" . $config->{'nix'}->{'nix_talker'} . "\" -- please make sure that you've got it set up and/or installed.\n";
	
		# Make it the default file...
		select SPEECH;
	
		# Automatically flush the buffer to it on every print...
		$| = 1;

		# Which Quake program are we using?
		init_quakeprog();

		# Set paths to key files...
		$agripdir_path = $base_path . 'id1';
		$autoexec_path = $agripdir_path . '/autoexec.cfg';
		$config_path = $agripdir_path . '/config.cfg';

		# Go to base path...
		# ONLY IF IN GAME MODE (otherwise we'd not be able to install QMODs
		# from different directories, because we'd switched away from the dir
		# that the .qmod is in)...
		chdir($base_path) unless $ARGV[0];

		# Do we need to send an init string?
		foreach my $key ( keys %{$config->{'nix_init_strings'}} )
		{
			print SPEECH $config->{'nix_init_strings'}->{$key} . "\n";
		}
	}
}

# Work out which quake program we're using and update $quake_prog accordingly...
sub init_quakeprog()
{
	my $engine = $config->{$os}->{'engine'};
	if( $engine =~ /gl/ )
	{
		u_update_quakeprog(RENDER_GL);
	}
	else
	{
		u_update_quakeprog(RENDER_SAFE);
	}
}


##############################
# Check we have the Quake data
##############################

sub datacheck()
{
	my $id1path = $base_path . 'id1' . $path_sep;
	my $answer;

	# Work out which data files we have...
	datacheck_detectpaks($id1path);

	# Error condition check -- can't have both pak1m and pak2m...
	u_die "ERROR: You appear to have two copies of the mindgrid pakfile.\nPlease contact the developers for help.\n" if $pak1m and $pak2m;

	# Move mindgrid to pak2.pak to avoid complications in the data
	# check and installation...
	if( $pak1m )
	{
		rename($id1path.'pak1.pak', $id1path.'pak2.pak');
		datacheck_detectpaks($id1path);
	}

	# Do we need to look for the game data?
	if( !$pak0 )
	{
		if( $os eq 'nix' )
		{
			# Un*x users should have already been given the option to install
			# the data by the setup script...
			u_die "Please install the shareware or registered Quake data.\nThe setup program can download the shareware version for you, or install the registered data from CD.\n";
		}
		else
		{
			# 'doze users may have the data on their hard drive somewhere...
			if( -r 'C:\QUAKE\id1\pak0.pak' )
			{
				# Shareware data was found...
				u_msg "Found the shareware Quake data on your computer.\nThis is required for AudioQuake to run.\nShould it be moved to your AudioQuake installation\n(press Y or N followed by ENTER)? ", 0;
				$answer = u_getanswer('y','n');
				if( $answer eq 'y' )
				{
					rename('C:\QUAKE\id1\pak0.pak', $id1path.'pak0.pak');
					
					# Look for registered data on fixed disk or CD...
					if( -r 'C:\QUAKE\id1\pak1.pak' )
					{
						rename('C:\QUAKE\id1\pak1.pak', $id1path.'pak1.pak');
						u_msg "The registered game data was found and moved as well.\n", 0;
					}
				}
			}
			else
			{
				# Shareware data not found on fixed disk; offer to install
				# from CD...
				u_msg "Neither the shareware or registered game data was found!\nIf you have a Registered Quake CD, the data can be installed from it now.\nIf not, you'll need to rerun the installer, making sure that you ask it\nto download and install the shareware version of Quake, as this is required for AudioQuake to run.\nWould you like to install the data from a CD now\n(press Y or N followed by ENTER) ", 0;
				$answer = u_getanswer('y', 'n');
				install_cdgamedata() if( $answer eq 'y' );
			}

			# We should now have at least the shareware data available...
			datacheck_detectpaks($id1path);
			u_die "AudioQuake cannot run without at least the shareware Quake data.\nPlease install the shareware or registered Quake data.\nYou can instruct the setup program to download the shareware demo for you.\n" if !$pak0;
		}
	}

	# This could be the first time the game was run.
	# If that's the case, we should ask the user if they want to find
	# the registered data on their hard disk or CD...
	if( $config->{'general'}->{'f_firstrun'} == 1 )
	{
		if( !$pak1r and $os eq 'win32' )
		{
			u_msg "The registered game data was not found.\nIt is not required to run AudioQuake, but if you have a Registered Quake CD,\nthe data can be installed from it now.\nIf not, you can use the shareware version.\nWould you like to install the data froma CD now?\nPlease enter y to do so or n to continue without.\n", 0;
			$answer = u_getanswer('y', 'n');
			install_cdgamedata() if( $answer eq 'y' );
		}
	}

	# Decide which version of the game and audio the user has...
	$registered = 1 if( $pak1r );
	$mindgrid = 1 if( $pak1m or $pak2m );

	# mindgrid must be pak1.pak to be loaded if we've just got the shareware
	# version...
	if( !$pak1r and $pak2m )
	{
		rename($id1path . 'pak2.pak', $id1path . 'pak1.pak');
	}
}

sub datacheck_detectpaks($)
{
	my $id1path = shift;

	# pak0.pak should be 18M (shareware)
	# pak1.pak should be 33M (registered)
	# pak2.pak should be 25M (mindgrid)
	
	# Initialise and detect files...
	$pak0 = $pak1r = $pak1m = $pak2m = 0;
	$pak0 = 1 if( -r $id1path . 'pak0.pak' );
	$pak1r = 1 if( -r $id1path . 'pak1.pak' and -s $id1path . 'pak1.pak' > 30_000_000 );
	$pak1m = 1 if( -r $id1path . 'pak1.pak' and -s $id1path . 'pak1.pak' < 30_000_000 );
	$pak2m = 1 if( -r $id1path . 'pak2.pak' );
	
	# Report what we found...
	#u_msg "pak0: $pak0, pak1r: $pak1r, pak1m: $pak1m, pak2m: $pak2m\n", 1;
}


##############################
# player_account
##############################

sub player_account()
{
	if( $config->{'general'}->{'f_firstrun'} == 1 )
	{
		u_msg "As this is the first time you've used this installation of AudioQuake,\nyou will need to provide your chosen username and password for playing online.\nIf you have not already registered a player name, you can register one\nat http://stats.agrip.org.uk/reg/new .  If you have registered a name already,\nyou can continue to use it and your current password.\nPlease register a player name on the Stats and Servers website before continuing\nIt is not a good idea to change your name as if you do you'll lose\nall stats associated with your old name.\nPlease press ENTER to continue.\n", 1;
		<STDIN>;
		player_account_update();
        if( $config->{'user'}->{'valid'} == 0 )
        {
            u_msg "Please ensure you have registered a player name\nat http://stats.agrip.org.uk/reg/new .\nYou can use the Update Name and Password Option on the Main Menu\nto correct this problem and allow you to play online games.\n", 0;
        }
	}
}

sub player_account_update()
{
	my $name;
	my $pass;
    my $exitflag = 0;

    u_msg "Please enter your chosen player name.\n", 1;
	$name = u_gettext();
	u_msg "Please enter your chosen password.\n", 1;
	$pass = u_gettext();

    if( u_authcheck($name,$pass) == 0 )
    {
        $config->{'user'}->{'valid'} = 1;
    }
    else
    {
        $config->{'user'}->{'valid'} = 0;
    }
	
	$config->{'user'}->{'name'} = $name;
	$config->{'user'}->{'pass'} = $pass;
	$config->write($base_path . 'start.ini');

	init_quakeprog();

    u_msg "Please press ENTER to continue.\n", 0;
    <STDIN>;
}


##############################
# main_menu
##############################

sub main_menu()
{
	my $choice;
	my $exitflag = 0;
	my $menutitle = "AudioQuake Launcher Main Menu.\n";
	my $menuheader;
	my $menutext;
	my $first_iteration = 1;

	if( $registered )
	{
		$menuheader = "Registered version";
	}
	else
	{
		$menuheader = "Shareware version";
	}
	if( $mindgrid )
	{
		$menuheader .= " with mindgrid audio.\n";
	}
	else
	{
		$menuheader .= " with standard audio.\n";
	}
	$menutext = "O, Accessibility options.\nK, Keybindings.\nM, Play a MOD.\nR, Rendering Options.\nS, Start.\nU, Update Name and Password.\nX, Exit.\nWhat is your choice?\n";

	while( ! $exitflag )
	{
		if( $first_iteration )
		{
			u_msg $menutitle . $menuheader . "Please choose an option by typing its letter and pressing ENTER.\n" . $menutext, 1;
			$first_iteration = 0;
		}
		else
		{
			u_msg $menutitle . $menutext, 1;
		}
		$choice = u_getanswer('o', 'k', 'm', 'r', 's', 'u', 'x');
	
		# Work out if we're to start the game or do something else...
		if( $choice eq 'o' )
		{
			main_accessopts();
		}
		elsif( $choice eq 'k' )
		{
			main_keybindings();
		}
		elsif( $choice eq 's' )
		{
			main_mainloop(LAUNCH_GAMEDIR,$config->{'general'}->{'default_gamedir'});
		}
		elsif( $choice eq 'm' )
		{
			main_modselector();
		}
		elsif( $choice eq 'r' )
		{
			main_renderopts();
		}
		elsif( $choice eq 'u' )
		{
			u_msg "You have selected to update your player name registration details.\nIf you change your name you'll lose all stats associated wiht your old name\nYou can track your progress using the website at http://stats.agrip.org.uk/ .\nPlease press ENTER to continue.\n", 0;
			<STDIN>;
			player_account_update();
		}
		elsif( $choice eq 'x' )
		{
			u_msg "Thanks for playing!\n", 1;
			if( $config->{'general'}->{'f_firstrun'} == 1 )
			{
				# Next run is not the first...
				$config->{'general'}->{'f_firstrun'} = 0;
				$config->write($base_path . 'start.ini');
			}
			$exitflag = 1;
		}
	}

	close SPEECH;
}


##############################
# main_mainloop
##############################

sub main_mainloop($$)
{
	# If the first paramater is LAUNCH_GAMEDIR, then
	#   open ZQ pipe specifying the +gamedir and the passed-in dta.
	# else
	#   open ZQ pipe specifying just what was passed in -- raw args (LAUNCH_RAW)
	# fi
	my $initflag = 0;
	my $line;
	my @prevlines;
	my $prevlines_ptr = -1;
	my $f_echo;
	my $f_keepline;
	my $mode = shift;
	my $gamedir_or_args = shift;

	$f_echo = $config->{'general'}->{'f_echo'};

    if( $config->{'user'}->{'valid'} != 1 )
    {
        u_msg "Either you have not registered a player name or the name and password you\nsupplied were not valid.  Please be aware that you will not be able to\nconnect to most games before resolving this problem.  You can use the\nUpdate Name and Password option on the main menu to fix this problem.\nThe game will now launch.\n", 1;
        <STDIN>;
    }
	
	# Open ZQuake with the specified parameters...
	if( $mode == LAUNCH_GAMEDIR )
	{
		open( INPUT, "$quake_prog +gamedir $gamedir_or_args|" ) or u_die "ERROR: Can't launch ZQuake!\n";
	}
	else
	{
		open( INPUT, "$quake_prog $gamedir_or_args|" ) or u_die "ERROR: Can't launch ZQuake!\n";
	}
	u_msg "Welcome to Quake.\n", 1;

	# Filter and Speak!
	while( <INPUT> )
	{
		if( !$_ )
		{
			u_die "An error has occured.  Please contact the developers and provide as much information on what the cause may be.  Thank you and apologie for the inconvenience.\n";
		}

		# FIXME need to ignore blank lines!

		$line = $_;
		chomp $line;
		next if length($line) == 0;
		$f_keepline = 0;  # assume we don't need to keep a record of this line.

		# Don't say owt unless the start text filter... 
		if( $initflag eq 0 )
		{
			$initflag = 1 if( $line =~ /ZQuake Initialized/ );
		}
		# We should filter output as normal...
		elsif( $line eq "TBo " )
		{
			# If this is the first time they've pressed the key, move older
			# from most recent line...
			if( $prevlines_ptr == -1 )
			{
				$prevlines_ptr = @prevlines;
			}
			
			# Move one message OLDER and set pointer...
			if( $prevlines_ptr > 0 )
			{
				$prevlines_ptr--;
				u_msg $prevlines_ptr-$#prevlines . ": " . $prevlines[$prevlines_ptr] . "\n", 1;
			}
		}
		elsif( $line eq "TBn " )
		{
			# If this is the first time they've pressed the key, they can't
			# move away from where they are as it's the latest item...
			if( $prevlines_ptr == -1 )
			{
				$prevlines_ptr = $#prevlines;
			}

			# Move one message NEWER and set pointer...
			if( $prevlines_ptr < $#prevlines )
			{
				$prevlines_ptr++;
				u_msg $prevlines_ptr-$#prevlines . ": " . $prevlines[$prevlines_ptr] . "\n", 1;
			}
		}
		elsif( $line eq "RLM " )
		{
			# Simply repeat the latest message -- don't alter our place in
			# the history...
			u_msg 'latest: ' . $prevlines[$#prevlines] . "\n", 1;
		}
		elsif( $line eq "STFU " )
		{
			# STFU Key...
			u_msg "\n", 1;
			# FIXME move to top of stack?
			$prevlines_ptr = @prevlines;
		}
		# The message is just one character...
		elsif( $f_echo and length($line) == 1 )
		{
			if( $line eq " " )
			{
				$line = "space";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "." )
			{
				$line = "dot";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "-" )
			{
				$line = "dash"; 
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "!" )
			{
				$line = "bang";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "`" )
			{
				$line = "tick";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "/" )
			{
				$line = "slash";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "\\" )
			{
				$line = "backslash";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "#" )
			{
				$line = "octathorpe";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "(" )
			{
				$line = "left paren";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq ")" )
			{
				$line = "right paren";
				u_msg $line . "\n", 1;
			}
			elsif( $line eq "," )
			{
				$line = "comma";
				u_msg $line . "\n", 1;
			}
			else
			{
				u_msg $line . "\n", 1;
			}
		}
		# The message is high-priority...
		elsif( $line =~ /^!/ )
		{
			$line =~ s/^!//;
			u_msg $line . "\n", 1;
			$f_keepline = 1 unless $line =~ /^(Entering|Exiting) console\.$/;
		}
		# The message is a normal priority message...
		else
		{
			$line =~ s/^]//;
			u_msg $line . "\n", 0;
			$f_keepline = 1;
		}

		# Record previous line...
		if( $f_keepline )
		{
			push @prevlines, $line;
		}
	}

	close INPUT;
}


##############################
# main_accessopts
##############################

sub main_accessopts()
{
	my @accessopts;	 # a list of hashes (one hash for each option)
	my $tmprec;		 # used to hold ref to anonymous hash
	my $inentry;		# flag - shows we found the start of an option entry
	my $foundentry;	 # flag - shows we found the start and then end of one
	my $answer;		 # action user wants to do in list/edit loop
	my $newval;		 # returned by editor
	my $optlistitem;	# The string that is built up for each opt in the list

	# This creates a list of access options.
	# Each record is a hash and has the form:
	#   * option	- the variable name
	#   * value	 - what it's set to
	#   * default   - default value for option
	#   * allowed   - a list of allowed values (OPTIONAL)
	#   * help	  - friendly name from last line of option in autoexec.cfg
	#   * longhelp  - the whole description from autoexec.cfg
	# The user can then review and change these key bindings.
	
	# There are different types of setting:
	#   * toggles (on/off)
	#   * modes (x/x/z) - allowed values are specified in config file
	#   * integer values (e.g. detection distances)
	#   * throttles (sound voluems; between 0 and 1)
	# The editing/validation procedure is slightly different for each type.

	# Key echo is not controlled by the engine, but the launcher.
	# However, it's not really fair to call it a launcher option as it only
	# surfaces during games.
	# Due to this, we make it look like it's a game option and put it at the
	# top of the list of accessibility options...
	$tmprec = {};
	$tmprec->{'option'} = 'launcher_echo';
	$tmprec->{'value'} = $config->{'general'}->{'f_echo'};
	$tmprec->{'longhelp'} = "This controls whether your keystrokes in the Quake console\nare announced.\n";
	$tmprec->{'help'} = 'keystroke echo (toggle)';
	$tmprec->{'default'} = 1;
	push @accessopts, $tmprec;

	# Get settings from the autoexec file...
	# Open the autoexec.cfg file...
	open( AUTOEXEC, $autoexec_path ) or u_die "ERROR: I couldn't open the autoexec.cfg file.\n";
	while( <AUTOEXEC> )
	{
		# Attempt to populate option records as we scan through file...
	  
		# start of option definition
		if( /^(?:set|server)(?:info)? (.+) (.+)/ )
		{
			$tmprec = {};
			$tmprec->{'option'} = $1;
			$tmprec->{'value'} = $2;

			$inentry = 1;
		}
		# optional specific values line (for mode-type options)
		elsif( /^\/\/ allowed values: ((\d+ ?)+)/ )
		{
			$tmprec->{'allowed'} = $1;
		}
		# lines inbetween start and end that are commented are added to the
		# long help text
		elsif( $inentry and /^\/\/ (?!default)(.+)/ )
		{
			$tmprec->{'longhelp'} .= $1 . "\n";
		}
		# end of option definition
		elsif( /^\/\/ default for (.+) is (\d+(\.\d+)?)/ )
		{
			$tmprec->{'help'} = $1;
			$tmprec->{'default'} = $2;

			# Finished filling in the details of this entry...
			$foundentry = 1;
			$inentry = 0;
		}

		# Add hash to array...
		if( $foundentry )
		{
			push @accessopts, $tmprec;
			$foundentry = 0;
		}
	}
	close AUTOEXEC;

	# Kick off the list/edit loop...
	u_msg "\nThe following is a list of your current accessibility options.\nTo advance through the list, press ENTER.\nTo edit an option, press E, followed by ENTER.\nFor help on an option, press H then ENTER.\nTo exit, press X follwed by ENTER.\nPress ENTER to continue.\n", 1;
	<STDIN>;

	# Go through list of options until the user forcibly exits...
	while( 1 )
	{
		# Take each option in turn...
		foreach my $href ( @accessopts )
		{
			# Stay on a given option until user wants to move off it
			# -- they may want to get help *and* edit in one session.
			while( 1 )
			{
				# Print the option name...
				$optlistitem = $href->{'help'} . " is " . $href->{'value'};
				$optlistitem .= " (default)" if $href->{'value'} eq $href->{'default'};
				$optlistitem .= " (help available)." if $href->{'longhelp'};
				u_msg $optlistitem . "\n", 1;
	
				# Skip to next / edit / help / exit...
				$answer = u_getanswer('nothing', 'e', 'h', 'x');
				if( $answer eq 'nothing' )
				{
					last;  # break out of this inner loop (for one option)
				}
				elsif( $answer eq 'e' )
				{
					$newval = main_accessopts_edit($href->{'option'}, $href->{'help'}, $href->{'allowed'});
					if( $newval ne $href->{'option'} )
					{
						main_accessopts_apply($href, $newval);
					}
				}
				elsif( $answer eq 'h' )
				{
					if( $href->{'longhelp'} )
					{
						u_msg $href->{'longhelp'}, 1;
					}
					else
					{
						u_msg "No extra help available for this option.\n", 1;
					}
					u_msg "Press ENTER to return to the option list.\n", 0;
					<STDIN>;
				}
				else # user said `x' -- exit this whole thing
				{
					return;
				}
			}
		}
		u_msg "End of list reached; press ENTER to return to top or X to exit.\n", 1;
		$answer = u_getanswer('nothing', 'x');
		return if( $answer eq 'x' );
	}
}


##############################
# main_accessopts_edit
##############################

sub main_accessopts_edit($$$)
{
	my $val = shift;
	my $help = shift;
	my $allowed = shift;
	my $success = 0;  # flag
	my $newval;  # temporary storage for new value entered by user

	# There are three types of options:
	#   toggles (1/0)
	#   modes (allowed set of values specified in cfg file)
	#   integers
	#   throttles (0-1)
	# Act according to which type of option we're editing.
   
	if( $help =~ /toggle/ )
	{
		u_msg "Please enter 1 to set this toggle or 0 to disable it.\n", 1;
		$newval = <STDIN>;
		chomp $newval;
		$success = 1 if( $newval eq 1 or $newval eq 0 )
	}
	elsif( $help =~ /mode/ )
	{
		u_msg "Allowed values for this setting are: $allowed.\nPlease enter one of them.\n", 1;
		$newval = u_getanswer(split(/ /, $allowed));
		$success = 1;
	}
	elsif( $help =~ /throttle/ )
	{
		u_msg "Please enter a value between zero and one.\n", 1;
		$newval = <STDIN>;
		chomp $newval;
		$success = 1 if( $newval ge 0 and $newval le 1 )
	}
	else
	{
		# (must be an integer)
		u_msg "Please enter a postive number for this option.\n", 1;
		$newval = <STDIN>;
		chomp $newval;
		$success = 1 if( $newval gt 0 and $newval !~ /[a-zA-Z]/ )
	}

	# Did we get valid input?	
	if( !$success )
	{
		$newval = "nothing" if ! $newval;
		u_msg "You entered $newval, which is invalid.\nPress ENTER to return to the options list without making changes.\n", 1;
		<STDIN>;
		# Keep old val setting...
		return $val;
	}
	else
	{
		return $newval;
	}
}


##############################
# main_accessopts_apply
##############################

sub main_accessopts_apply($$)
{
	my $href = shift;
	my $newval = shift;

	# Most options are set in autoexec.cfg, but one (keystroke echo) is not...
	if( $href->{'option'} ne 'launcher_echo' )
	{
		# Open the real one and a temporary one...
		open( AUTOEXEC, $autoexec_path ) or u_die "ERROR: I couldn't open the autoexec.cfg file to save your settings.\n";
		open( AUTOETMP, ">$autoexec_path.tmp" ) or u_die "ERROR: I couldn't open the autoexec.cfg.tmp file to save your settings.\n";
	
		# Echo most lines back, updating the one that has just been changed...
		while( <AUTOEXEC> )
		{
			if( /^(?:set|server)(?:info)? $href->{'option'} $href->{'value'}/ )
			{
				my $line = $_;
				$line =~ s/$href->{'value'}/$newval/;
				print AUTOETMP $line;
			}
			else
			{
				print AUTOETMP;
			}
		}
	
		close AUTOEXEC;
		close AUTOETMP;
	
		rename "$autoexec_path.tmp", $autoexec_path;

		u_update_autoexec();
	}
	else
	{
		# Update keystroke echo config variable...
		$config->{'general'}->{'f_echo'} = $newval;
		$config->write($base_path.'start.ini');
	}

	# Update the option hash so that the new value is reflected in the list...
	$href->{'value'} = $newval;
}


##############################
# main_keybindings
##############################

sub main_keybindings()
{
	my @keybindings;
	my $tmprec;		 # used to hold ref to anonymous hash
	my $foundentry;	 # flag
	my $answer;		 # action user wants to do in list/edit loop
	my $newkey;		 # returned by editor

	# This creates a list of keybindings.
	# Each record is a hash and has the form:
	#   * key
	#   * action
	#   * file (autoexec or config)
	#   * default (default key for action)
	#   * help text (from comment in autoexec.cfg)
	# The user can then review and change these key bindings.

	# Get bindings from the autoexec file...
	# Open the autoexec.cfg file...
	open( AUTOEXEC, $autoexec_path ) or u_die "ERROR: I couldn't open the autoexec.cfg file.\n";
	while( <AUTOEXEC> )
	{
		# Get each line of the file that is a key binding and then the
		# next line that has the format of a key binding description...
		
		if( /bind "(.+)" "(.+)"/ )
		{
			# Flag and create a new anonymous hash...
			$foundentry = 1;
			$tmprec = {};

			# Start to fill hash...
			$tmprec->{'key'} = $1;
			$tmprec->{'action'} = $2;
			$tmprec->{'file'} = 'autoexec';
		}

		if( /\/\/ default key (to|for) (.+) is "(.+)"/ )
		{
			$tmprec->{'default'} = $3;
			$tmprec->{'help'} = $2;
		}

		# Add hash to array...
		if( $foundentry )
		{
			push @keybindings, $tmprec;
			$foundentry = 0;
		}
	}
	close AUTOEXEC;

	# Now get the bindings from the config file...
	# Open the config.cfg file...
	open( CONFIG, $config_path ) or u_die "ERROR: I couldn't open the config.cfg file.\n";
	while( <CONFIG> )
	{
		# Get each line of the file that is a key binding and then the
		# next line that has the format of a key binding description...
		
		if( /bind (.+) "[+-]?(.+)"/ and ( $2 !~ /aga/ ) )
		{
			$tmprec = {};
		
			$tmprec->{'key'} = $1;
			$tmprec->{'action'} = $2;
			$tmprec->{'file'} = 'config';

			# Do some special filtering to make help messages more, erm, helpful...
			$tmprec->{'key'} = '`' if( $tmprec->{'key'} eq 'backquote' );
			$tmprec->{'help'} = $tmprec->{'action'};
			$tmprec->{'help'} =~ s/impulse/switch to weapon/;
			$tmprec->{'help'} =~ s/menu_(.+)/show the $1 menu/ if( $tmprec->{'help'} =~ /menu_/ );
			$tmprec->{'help'} = 'show the help grahpics' if( $tmprec->{'help'} eq 'help' );
			$tmprec->{'help'} = 'Quickload' if( $tmprec->{'help'} =~ /uickload/ );
			$tmprec->{'help'} = 'Quicksave' if( $tmprec->{'help'} =~ /uicksav/ );
			$tmprec->{'help'} =~ tr/_/ /;
	   
			# Push ref to anonymous hash onto list...
			push @keybindings, $tmprec;
		}
	}
	close CONFIG;

	# Kick off the list/edit loop...
	u_msg "\nThe following is a list of your current keybindings.\nTo advance through the list, press ENTER.\nTo edit a binding, press E, followed by ENTER.\nTo exit, press X follwed by ENTER.\nPress ENTER to continue.\n", 1;
	<STDIN>;

	# Loop through each action bound to a key...
	while( 1 )
	{
		for my $href ( @keybindings )
		{
			# Print the action name...
			u_msg $href->{'help'} . " = " . $href->{'key'}, 1;
			if( $href->{'default'} and ( $href->{'key'} eq $href->{'default'} ) )
			{
				u_msg " (default)\n", 0;
			}
			else
			{
				u_msg "\n", 0;
			}

			# Skip to next / edit / exit...
			$answer = u_getanswer('nothing', 'e', 'x');
			if( $answer eq 'nothing' )
			{
				next;
			}
			elsif( $answer eq 'e' )
			{
				$newkey = main_keybindings_edit($href->{'key'}, $href->{'help'});
				if( $newkey ne $href->{'key'} )
				{
					main_keybindings_apply($href, $newkey);
				}
			}
			else
			{
				return;
			}
		}
		u_msg "End of list reached; press ENTER to return to top or X to exit.\n", 1;
		$answer = u_getanswer('nothing', 'x');
		return if( $answer eq 'x' );
	}
}


##############################
# main_keybindings_edit
##############################

sub main_keybindings_edit($$)
{
	my $key = shift;
	my $help = shift;
	
	u_msg "Please enter the new key for the action:\n$help.\nSpecial keys should be typed with their long names (e.g. enter or f10).\n", 1;

	my $newkey = <STDIN>;
	chomp $newkey;
	$newkey =~ tr/A-Z/a-z/;

	if( $newkey =~ /[a-z0-9]+/ )
	{
		return $newkey;
	}
	else
	{
		u_msg "You entered $newkey, which is invalid.\nReturning to key bindings list without making changes.\n", 1;
		# Keep old key setting...
		return $key;
	}
}


##############################
# main_keybindings_apply
##############################

sub main_keybindings_apply($$)
{
	my $href = shift;
	my $newkey = shift;

	# The key could be in config.cfg or autoexec.cfg.
	# This alters significantly how we should behave...
	
	if( $href->{'file'} eq "config" )
	{
		# config.cfg...

		# Open the config.cfg file and a temporary file...
		open( CONFIG, $config_path ) or u_die "ERROR: I couldn't open the config.cfg file to save your setting.\n";
		open( CFGTMP, ">$config_path.tmp" ) or u_die "ERROR: I couldn't open the config.cfg.tmp file to save your setting.\n";
	
		# Echo each line of the file to a temporary file, updating this binding...
		while( <CONFIG> )
		{
			# Search for the binding corresponding to the one we wish to update...
			if( /bind $href->{'key'} "[+-]?$href->{'action'}"/ )
			{
				my $line = $_;
				$newkey =~ tr/a-z/A-Z/;
				$line =~ s/$href->{'key'}/$newkey/;
				print CFGTMP $line;
			}
			else
			{
				print CFGTMP;
			}
		}

		close CONFIG;
		close CFGTMP;

		rename "$config_path.tmp", $config_path;

		u_update_config();
	}
	else
	{
		# autoexec.cfg...

		# Open the real one and a temporary one...
		open( AUTOEXEC, $autoexec_path ) or u_die "ERROR: I couldn't open the autoexec.cfg file to save your settings.\n";
		open( AUTOETMP, ">$autoexec_path.tmp" ) or u_die "ERROR: I couldn't open the autoexec.cfg.tmp file to save your settings.\n";

		# Echo most lines back, updating the one that has just been changed...
		while( <AUTOEXEC> )
		{
			if( /bind "$href->{'key'}"/ )
			{
				my $line = $_;
				$line =~ s/$href->{'key'}/$newkey/;
				print AUTOETMP $line;
			}
			else
			{
				print AUTOETMP;
			}
		}

		close AUTOEXEC;
		close AUTOETMP;

		rename "$autoexec_path.tmp", $autoexec_path;

		u_update_autoexec();
	}
}


##############################
# ModSelector
##############################

sub main_modselector()
{
	my @modlist;  # a list of hashes (each one stores info on won emod)
	my $thismodprompt;  # a string built up to prompt user of each mod found.
	my $answer;

	# We will search for mods and put their details into a list that can be
	# reviewed by the player, just as the access options and keybindings
	# editing menus work above.
	#
	# Each record in the aray is a hash, with:
	#   * name
	#   * version
	#   * shortdesc
	#   * longdesc (if provided by mod author)
	#   * gamedir
	#
	# This information is used for pretty-printed output for the dear user.

	@modlist = main_modselector_scan();


	# Kick off the list/edit loop (or exit)...
	if( @modlist > 0 )
	{
		u_msg "\nThe following is a list of mods you have installed.\nTo advance through the list, press ENTER.\nTo get a longer description of a mod, press D, then ENTER.\nTo play the game using this mod, press P and ENTER.\nRemove the mod by typing R and ENTER.\nTo exit, press X follwed by ENTER.\nPress ENTER to continue.\n", 1;
		<STDIN>;
	}

	# Go through list of mods until the user forcibly exits...
	while( @modlist > 0 )
	{
		# Take each option in turn...
		foreach my $href ( @modlist )
		{
			# Stay on a given mod until user wants to move off it
			# -- they may want to get help *and* edit in one session.
			while( 1 )
			{
				# Print the vital stats...
				$thismodprompt = $href->{'name'} . ", " . $href->{'shortdesc'};
				$thismodprompt .= " (long description available)." if $href->{'longdesc'};
				$thismodprompt .= "\n";
				u_msg $thismodprompt, 1;
	
				# Skip to next / edit / help / exit...
				$answer = u_getanswer('nothing', 'd', 'p', 'r', 'x');
				if( $answer eq 'nothing' )
				{
					last;  # break out of this inner loop (for one option)
				}
				elsif( $answer eq 'p' )
				{
					main_mainloop(LAUNCH_GAMEDIR,$href->{'gamedir'});
				}
				elsif( $answer eq 'd' )
				{
					if( $href->{'longdesc'} )
					{
						u_msg $href->{'longdesc'}, 1;
					}
					else
					{
						u_msg "No long description available for this mod.\n", 1;
					}
					u_msg "Press ENTER to return to the option list.\n", 0;
					<STDIN>;
				}
				elsif( $answer eq 'r' )
				{
					qmod_shreddir $base_path . $href->{'gamedir'};
					u_msg "This mod has been removed.\nPress ENTER to continue.\n", 1;
					@modlist = main_modselector_scan();
					<STDIN>;
					last;
				}
				else # user said `x' -- exit this whole thing
				{
					return;
				}
			}
		}
		u_msg "End of list reached; press ENTER to return to top or X to exit.\n", 1;
		$answer = u_getanswer('nothing', 'x');
		return if( $answer eq 'x' );
	}

	# Why do we only do this here?  Well it could be that we remove a mod in the
	# loop above, so we need to catch that exit here...
	u_msg "There are no mods currently installed.\nRead the manual and web site for information on mods and how to get them.\nPress ENTER to return to the main menu.\n", 1;
	<STDIN>;
	return;
}


##############################
# ModSelector Scanning Routine
##############################

sub main_modselector_scan()
{
	my @gamedirs;
	my @mods;
	my $longdesc;
	my $tmprec;

	# Find mods...
	@gamedirs = qmod_listmods();
	foreach my $gamedir ( @gamedirs )
	{
		# No error-checking here -- should have been done on install...

		my $modconf = Config::Tiny->read($gamedir . $path_sep . 'qmod.ini');
		$tmprec = {};
		$tmprec->{'name'} = $modconf->{'general'}->{'name'};
		$tmprec->{'version'} = $modconf->{'general'}->{'version'};
		$tmprec->{'shortdesc'} = $modconf->{'general'}->{'shortdesc'};
		$tmprec->{'gamedir'} = $modconf->{'general'}->{'gamedir'};

		# Get longdesc if it's there...
		if( $modconf->{'longdesc'} )
		{
			foreach my $key ( sort keys %{$modconf->{'longdesc'}} )
			{
				$longdesc .= $modconf->{'longdesc'}->{$key} . "\n";
			}
			$tmprec->{'longdesc'} = $longdesc;
		}

		push @mods, $tmprec;
	}

	return @mods;
}


##############################
# Rendering Options
##############################

sub main_renderopts()
{
	my $exitflag;
	my $choice;
	my $menutext;

	while( ! $exitflag )
	{
		# Construct menu depending on OS...
		$menutext = "Graphics Rendering Options.\nThis menu allows you to enable graphical output.\nPlease select an option from the list below.\n";
		if( $os eq 'win32' )
		{
			$menutext .= "N, no graphical output.\nO, OpenGL graphical output.\n";
		}
		else
		{
			$menutext .= "S, Software rendering.\nG, OpenGL rendering.\n";
		}
		$menutext .= "X, Return to main menu without making a change.\nWhich type of output would you like?\n";
		u_msg $menutext, 1;

		# Get the user's answer -- valid answers depend on OS...
		if( $os eq 'win32' )
		{
			$choice = u_getanswer('n', 'o', 'x');
		}
		else
		{
			$choice = u_getanswer('s', 'g', 'x');
		}
	
		# Act on the user's choice...
		if( $choice eq 'n' )
		{
			$config->{'win32'}->{'engine'} = 'zquake-vidnull.exe';
			$config->write($base_path . 'start.ini');
			u_update_quakeprog(RENDER_SAFE);
			u_msg "Settings updated. Press ENTER to return to the main menu.\n", 1;
			<STDIN>;
			$exitflag = 1;
		}
		elsif( $choice eq 'o' )
		{
			$config->{'win32'}->{'engine'} = 'zquake-gl.exe';
			$config->write($base_path . 'start.ini');
			u_update_quakeprog(RENDER_GL);
			u_msg "Settings updated. Press ENTER to return to the main menu.\n", 1;
			<STDIN>;
			$exitflag = 1;
		}
		elsif( $choice eq 's' )
		{
			$config->{'nix'}->{'engine'} = 'zquake-sdl';
			$config->write($base_path . 'start.ini');
			u_update_quakeprog(RENDER_SAFE);
			u_msg "Settings updated. Press ENTER to return to the main menu.\n", 1;
			<STDIN>;
			$exitflag = 1;
		}
		elsif( $choice eq 'g' )
		{
			$config->{'nix'}->{'engine'} = 'zquake-glsdl';
			$config->write($base_path . 'start.ini');
			u_update_quakeprog(RENDER_GL);
			u_msg "Settings updated. Press ENTER to return to the main menu.\n", 1;
			<STDIN>;
			$exitflag = 1;
		}
		elsif( $choice eq 'x' )
		{
			$exitflag = 1;
		}
	}
}


##############################
# Registered data installer
##############################

# Main installer function...
sub install_cdgamedata()
{
	my $answer;
	my $drive;

	# Install data from CD...
	u_msg "Installing data from CD...\n", 1;
   
	# Have we got lha?
	if( ! -e "lha.exe" )
	{
		u_msg "ERROR: the lha.exe program cannot be found.  It is required to install data\nfrom your Quake CD. Please reinstall AudioQuake to replace this file\nand rerun this program.  Exiting...\n", 1;
		exit;
	}

	# This will call itself and examinedisc until we've got things right...
	$drive = install_getdrive(); 

	# Found files, begin the install...
	u_msg "All files on CD accounted for; beginning the installation process...\n", 1;
	u_msg "1. Getting compressed files from the CD...\n", 0;
	if( $AS_Q106 )
	{
		system("copy $drive:\\" . AS_QARC_FULLNAME . " .");
	}
	else
	{
		system("copy /B $drive:\\" . AS_QARC_PT1 . "+$drive:\\" . AS_QARC_PT2 . '  ' . AS_QARC_FULLNAME);
	}
	u_msg "2. Extracting data files...\n", 0;
	system("lha.exe xq " . AS_QARC_FULLNAME);
	u_msg "3. Cleaning up...\n", 0;
	system("del " . AS_QARC_FULLNAME);

	# Install complete!
	u_msg "Installation complete.  Press ENTER to go to the main menu.\n", 0;
	<STDIN>;
}

# CD-checking support function for install...
sub install_examinedisc($)
{
	my $drive = $_[0];
	my $retval = -1;

	u_msg "Checking for Quake CD in drive $drive...\n", 1;

	# Work out CD version...
	if( -e "$drive:\\" . AS_QARC_FULLNAME )
	{
		$AS_Q106 = 1;
		$retval = 1;
	}
	else
	{
		if( ! -e "$drive:\\" . AS_QARC_PT1 )
		{
			u_msg "ERROR: I cannot find the " . AS_QARC_PT1 . " file on the CD.\n", 0;
		}
		elsif ( ! -e "$drive:\\" . AS_QARC_PT2 )
		{
			u_msg "ERROR: I cannot find the " . AS_QARC_PT2 . " file on the CD.\n", 0;
		}
		else
		{
			$retval = 1;
		}
	}

	return $retval;
}

# Function to get a sane Quake CD drive...
sub install_getdrive()
{
	my $drive;
	my $exitflag = 0;

	u_msg "Please enter the letter of the drive in which your Quake CD can be found.\n", 0;
	while( ! $exitflag )
	{
		$drive = <STDIN>;
		chomp $drive;
		if( $drive =~ /[a-z]/ )
		{
			if( install_examinedisc($drive) > 0 )
			{
				$exitflag = 1;
			}
			else
			{
				u_msg "It doesn't appear that you've inserted a Quake CD in drive $drive.\nPlease ensure you've inserted a Quake CD and enter the letter of the drive it is in.\n", 0;
			}
		}
		else
		{
			u_msg "You didn't enter a valid drive letter.  Please enter a letter between A nd Z.\n", 1;
		}
	}
	
	return $drive;
}


##############################
# QMOD Main Logic
##############################

sub qmod_main($)
{
	my $qmodfile = shift;
	my $qmodconf;
	my $ex_qmodconf;
	my $unzip;
	my $ini_err_msg;
	my $tempdir;
	my $answer;

	# Sanity Checks...

	# Find the unzip binary...
	if( $os eq 'nix' )
	{
		$unzip = '/usr/bin/unzip' if -x '/usr/bin/unzip';
		u_die "I can't find an unzip program on your system - aborting QMOD installation.\n" if !$unzip;

		$tempdir = '/tmp/audioquake-mod-install/';
	}
	else
	{
		$unzip = 'unzip.exe' if -x 'unzip.exe';
		u_die "I can't find the unzip program included with AudioQuake.\nIt should have been installed with the game.\n Please contact the developers.\nAborting QMOD installation.\n" if !$unzip;

		$tempdir = $base_path . 'temp\\';
	}

	# A .qmod file is really a ZIP file that contains the gamedir of the mod
	# itself.
	# 
	# A file called qmod.ini should end up in the AQ directory.  This is
	# scanned to find out all sorts of information about the mod.
	#
	# One thing in qmod.ini is the gamedir's name (this allows us to have .qmod
	# files with different names to the mod's gamedir -- so that, for example,
	# the version of the mod can be incorporated into the .qmod filename).
	# 
	# [ Please read the development manual for more detailed information! ]
	#
	# Once the qmod.ini has been examined, it is moved to the gamedir of the
	# modification.

	# Extract to a temporary dir, then analyse INI file...
	u_msg "QMOD Installer.  Unpacking files...\n", 1;
	system($unzip,$qmodfile,'-d',$tempdir);

	# Examine the qmod.ini file...
	$qmodconf = Config::Tiny->read($tempdir . 'qmod.ini') or u_die "ERROR: The author of " . $qmodfile . " has not included a qmod.ini file.\nThis is needed to complete the installation.\nPlease remove the modification's files from your AudioQuake directory and\nreport this as a bug to the modification's author.\n";

	# Check for required fields...
	$ini_err_msg = "ERROR: The author of " . $qmodfile . " has not included some required information\nin the modification's qmod.ini file.\nThis is needed to complete the installation.\nPlease remove the modification's files from your AudioQuake directory and\nreport this as a bug to the modification's author.\n";
	u_die $ini_err_msg unless $qmodconf->{'general'}->{'name'};
	u_die $ini_err_msg unless $qmodconf->{'general'}->{'version'};
	u_die $ini_err_msg unless $qmodconf->{'general'}->{'shortdesc'};
	u_die $ini_err_msg unless $qmodconf->{'general'}->{'gamedir'};
	u_die $ini_err_msg unless $qmodconf->{'general'}->{'watch_config'};
	u_die $ini_err_msg unless $qmodconf->{'general'}->{'watch_autoexec'};

	u_msg $qmodconf->{'general'}->{'name'} . " " . $qmodconf->{'general'}->{'version'} . " installation.\n", 0;

	# Do we already have a copy of the mod?
	if( -d $base_path . $qmodconf->{'general'}->{'gamedir'} )
	{
		u_msg "You already have a mod installed in the directory the new one wishes to use.\n", 0;
		# See if we can get any information about the existing mod...
		$ex_qmodconf = Config::Tiny->read($base_path . $qmodconf->{'general'}->{'gamedir'} . $path_sep . 'qmod.ini');
		if( $ex_qmodconf )
		{
			u_msg "Some information on the existing mod:\n", 0;
			u_msg "Name: " . ( $ex_qmodconf->{'general'}->{'name'} ? $ex_qmodconf->{'general'}->{'name'} : "(information not found)" ) . "\n", 0;
			u_msg "Version: " . ( $ex_qmodconf->{'general'}->{'version'} ? $ex_qmodconf->{'general'}->{'version'} : "(information not found)" ) . "\n", 0;
			u_msg "Short description: " . ( $ex_qmodconf->{'general'}->{'shortdesc'} ? $ex_qmodconf->{'general'}->{'shortdesc'} : "(information not found)" ) . "\n", 0;
		}
		else
		{
			u_msg "No information about the existing mod can be found.\nYou may want to examine the situation yourself before proceeding.\n", 0;
		}

		# Ask if user want's to replace existing mod with new one...
		u_msg "Would you like to remove the existing mod and replace it with the new one?\n", 0;
		$answer = u_getanswer('y','n');
		if( $answer eq 'n' )
		{
			qmod_shreddir($tempdir);
			u_die "Installation Aborted.\n";
		}
		u_msg "\n", 1;
	}

	# Move the gamedir from the temporary directory to $base_path...
	rename($tempdir . $qmodconf->{'general'}->{'gamedir'}, $base_path . $qmodconf->{'general'}->{'gamedir'});
 
	# Move qmod.ini to the gamedir...
	rename($tempdir . 'qmod.ini',$base_path . $qmodconf->{'general'}->{'gamedir'} . $path_sep . 'qmod.ini');

	# Now remove the temporary directory.
	qmod_shreddir $tempdir or u_die "ERROR: Can't remove temporary directory!\n";

	# Copy across autoexec and config files?
	if( qmod_ccquery_config $qmodconf->{'general'}->{'gamedir'} )
	{
		u_msg "Installing config.cfg...\n", 0;
		u_filecopy $agripdir_path . $path_sep . 'config.cfg', $base_path . $qmodconf->{'general'}->{'gamedir'} . $path_sep . 'config.cfg';
	}
	if( qmod_ccquery_autoexec $qmodconf->{'general'}->{'gamedir'} )
	{
		u_msg "Installing autoexec.cfg...\n", 0;
		u_filecopy $agripdir_path . $path_sep . 'autoexec.cfg', $base_path . $qmodconf->{'general'}->{'gamedir'} . $path_sep . 'autoexec.cfg';
	}

	# Put longdesc if there...
	u_msg "Here is some more information about the mod...\n", 0;
	if( $qmodconf->{'longdesc'} )
	{
		foreach my $key ( sort keys %{$qmodconf->{'longdesc'}} )
		{
			u_msg $qmodconf->{'longdesc'}->{$key} . "\n", 0;
		}
	}
	else
	{
		u_msg $qmodconf->{'general'}->{'shortdesc'} . "\n", 0;
	}

	# There is no step three!
	u_msg "Installation complete.\nPress ENTER to exit.\n", 0;
	<STDIN>;
}


##############################
# Recursive directory destroyer
##############################

sub qmod_shreddir($)
{
	# This function recursively removes all files in a directory.
	# It is DANGEROUS to use it, but we need it.
	my $dir = shift;
	my $retval;
	return -1 if !-d $dir;
	chdir $dir;
	my @contents = <*>;
	#u_msg "DIR $dir CONTENTS: @contents\n", 1;
	foreach my $item ( @contents )
	{
		if( -d $item )
		{
			$retval = qmod_shreddir($item);
			#u_msg "RETVAL $retval\n", 1;
		}
		else
		{
			#u_msg "UNLINK $item!\n", 1;
			unlink $item or return 0;
		}
	}
	chdir '..';
	#u_msg "RMDIR $dir!\n", 1;
	rmdir $dir or return 0;
	return 1;
}


##############################
# MOD Lister
##############################

sub qmod_listmods()
{
	my @fsobjs;
	my @gamedirs;

	# We return all MODs in the base directory.
	# This is run from $base_path.
	
	@fsobjs = <*>;
	foreach my $fsobj ( @fsobjs )
	{
		if( -d $fsobj and -r $fsobj . $path_sep . 'qmod.ini' )
		{
			push @gamedirs, $fsobj;
		}
	}

	return @gamedirs;
}


##############################
# Return gamedirs of mods wanting updates to config.cfg
##############################

sub qmod_cclist_config()
{
	my @gamedirs;
	my @config_watchers;
	my $modconf;

	@gamedirs = qmod_listmods();

	foreach my $gamedir ( @gamedirs )
	{
		# Find out if mod wants to be updated with config.cfg changes...
		$modconf = Config::Tiny->read($base_path . $gamedir . $path_sep . 'qmod.ini');
		push @config_watchers, $gamedir if $modconf->{'general'}->{'watch_config'} eq 'yes';
	}

	return @config_watchers;
}


##############################
# Query if a specific mod requires config updates
##############################

sub qmod_ccquery_config($)
{
	# Find out if mod wants to be updated with config.cfg changes...
	# NOTE: We assume the gamedir is valid!
	my $gamedir = shift;
	my $modconf = Config::Tiny->read($base_path . $gamedir . $path_sep . 'qmod.ini');
	return 1 if $modconf->{'general'}->{'watch_config'} eq 'yes';
	return 0;
}


##############################
# Return gamedirs of mods wanting updates to autoexec.cfg
##############################

sub qmod_cclist_autoexec()
{
	my @gamedirs;
	my @autoexec_watchers;
	my $modconf;

	@gamedirs = qmod_listmods();

	foreach my $gamedir ( @gamedirs )
	{
		# Find out if mod wants to be updated with config.cfg changes...
		$modconf = Config::Tiny->read($gamedir . $path_sep . 'qmod.ini');
		push @autoexec_watchers, $gamedir if $modconf->{'general'}->{'watch_autoexec'} eq 'yes';
	}

	return @autoexec_watchers;
}


##############################
# Query if a specific mod requires autoexec updates
##############################

sub qmod_ccquery_autoexec($)
{
	# Find out if mod wants to be updated with config.cfg changes...
	# NOTE: We assume the gamedir is valid!
	my $gamedir = shift;
	my $modconf = Config::Tiny->read($base_path . $gamedir . $path_sep . 'qmod.ini');
	return 1 if $modconf->{'general'}->{'watch_autoexec'} eq 'yes';
	return 0;
}


##############################
# Utility Functions
##############################

# Print out a u_msg to STDOUT and SPEECH...
sub u_msg($$)
{
	my $message = shift;
	my $highpri_flag = shift;
	
	print STDOUT $message if $config->{'general'}->{'f_stdout'};

	# Send the clear buffer command/prefix if it's a high priority message.
	if( $highpri_flag )
	{
		if( $os eq 'nix' )
		{
			print SPEECH $config->{'nix'}->{'nix_highpri_cmd'} . "\n";
		}
		else
		{
			$message = $config->{'win32'}->{'win32_highpri_prefix'} . $message;
		}
	}

	# Now send the text to be spoken itself...
	if( $os eq 'nix' )
	{
		print SPEECH $config->{'nix'}->{'nix_prefix'} . $message . $config->{'nix'}->{'nix_suffix'} . "\n";
	}
	else
	{
		print SPEECH $message;
	}
}

# Print out a u_msg to STDOUT and SPEECH, wait for ENTER, then exit...
sub u_die($)
{
	my $message = shift;
	u_msg $message . "  Press ENTER to exit...\n", 1;
	<STDIN>;
	die;
}

# Get a valid choice out of a number of possible answers...
sub u_getanswer(@)
{
	my @possible_answers = @_;
	my $answer;
	my $flag = 0;
	my $element;
	my $i = 0;

	until( $flag == 1 )
	{
		$answer = <STDIN>;
		chomp $answer;
		$answer =~ tr/[A-Z]/[a-z]/;

		# <> may be a valid answer, in which case it's been passed
		# to this function as ``nothing''...
		$answer = "nothing" if !$answer;

		# Have we got a valid answer?
		if( $answer )
		{
			foreach $element ( @possible_answers )
			{
				if( $answer eq $element )
				{
					$flag = 1;
					last;
				}
			}
		}

		# If we haven't, print out an appropriate error message and get the
		# user to try again...
		if( $flag == 0 )
		{
			u_msg "You entered $answer.\nPossible choices are ", 1;
			while( $i < $#possible_answers )
			{
				u_msg "$possible_answers[$i], ", 0;
				$i++;
			}
			u_msg "and $possible_answers[$#possible_answers]. ", 0;
			u_msg "Please try again.\n", 0;
			$i = 0;
		}
	}

	return $answer;
}

# Get some free text from the user...
sub u_gettext()
{
	my $answer;
	my $flag = 0;

	until( $flag == 1 )
	{
		$answer = <STDIN>;
		chomp $answer;
        #$answer =~ tr/[A-Z]/[a-z]/; can't do this or pwd will be altered and not validate

		# Have we got a valid answer?
		if( $answer )
		{
			$flag = 1;
		}
		else
		{
			$answer = 'nothing';
		}

		# If we haven't, print out an appropriate error message and get the
		# user to try again...
		if( $flag == 0 )
		{
			u_msg "You entered $answer.\nThis is invalid.\n", 1;
			u_msg "Please try again.\n", 0;
		}
	}

	return $answer;
}

# To avoid us having to install a module just to copy text files...
sub u_filecopy($$)
{
	my $src = shift;
	my $dest = shift;

	open SRC, '<' . $src or u_die "ERROR: Can't open " . $src . " for reading!\n";
	open DEST, '>' . $dest or u_die "ERROR: Can't open " . $dest . " for writing!\n";
	while( <SRC> )
	{
		print DEST;
	}
	close DEST;
	close SRC;
}

# Update mods watching config.cfg...
sub u_update_config()
{
	my @mods;
	my $qmodconf;
	@mods = qmod_cclist_config();
	foreach my $mod ( @mods )
	{
		$qmodconf = Config::Tiny->read($base_path . $mod . $path_sep . 'qmod.ini');
		if( $qmodconf->{'general'}->{'gamedir'} )
		{
			u_msg "Installing $mod config.cfg...\n", 0;
			u_filecopy $agripdir_path . $path_sep . 'config.cfg', $base_path . $qmodconf->{'general'}->{'gamedir'} . $path_sep . 'config.cfg';
		}
	}
}

# Update mods watching autoexec.cfg...
sub u_update_autoexec()
{
	my @mods;
	my $qmodconf;
	@mods = qmod_cclist_config();
	foreach my $mod ( @mods )
	{
		$qmodconf = Config::Tiny->read($base_path . $mod . $path_sep . 'qmod.ini');
		if( $qmodconf->{'general'}->{'gamedir'} )
		{
			u_msg "Installing $mod autoexec.cfg...\n", 0;
			u_filecopy $agripdir_path . $path_sep . 'autoexec.cfg', $base_path . $qmodconf->{'general'}->{'gamedir'} . $path_sep . 'autoexec.cfg';
		}
	}
}

# Update the $quake_prog variable...
sub u_update_quakeprog($)
{
	# Pass in RENDER_SAFE to indicate VN/sw rendering; RENDER_GL to indicate GL rendering.
	my $mode = shift;
	my $userpass = u_get_user_pass();

	if( $os eq 'win32' )
	{
		if( $mode == RENDER_SAFE )
		{
			$quake_prog = WIN32_ENGINE_OPTS_PRE . $base_path . $config->{'win32'}->{'engine'}  . WIN32_ENGINE_OPTS_VN;
		}
		else
		{
			$quake_prog = WIN32_ENGINE_OPTS_PRE . $base_path . $config->{'win32'}->{'engine'}  . WIN32_ENGINE_OPTS_GL;
		}
	}
	else
	{
		$quake_prog = $base_path . $config->{'nix'}->{'engine'} . NIX_ENGINE_OPTS_STD;
	}

	$quake_prog .= " " . $userpass;
}

# Turn the user and pass into a string for use on the command line...
sub u_get_user_pass()
{
	return "+name " . $config->{'user'}->{'name'} . " +mauth_password " . $config->{'user'}->{'pass'};
}

# Check if supplied auth details are right...
sub u_authcheck($$)
{
    # Returns:
    #   0   OK
    #   1   BAD
    #   127 ERROR
	
    use LWP::UserAgent;
	use HTTP::Request::Common;
	
	my $url = "http://stats.agrip.org.uk/reg/check/";
	my $name = "AGRIP AudioQuake Credentials Checker Script";
	my $page;
	my %fields;
	my $player_name = shift;
	my $player_pass = shift;
	
	%fields = (
	   "playername" => $player_name,
	   "playerpass" => $player_pass
	);

    my $ua = new LWP::UserAgent;
	$ua->agent($name);
	$page = $ua->request(POST $url,\%fields);
	
	if ($page->is_success)
	{
		if( $page->content =~ /Check Successful/ )
		{
            u_msg "Your supplied player name and password are correct.\nYou can play online games using them.\n", 0;
			return 0;
		}
		else
		{
            u_msg "Your supplied player name and password are incorrect.\nYou cannot play online games using them.\n", 0;
			return 1;
		}
	}
	else
	{
        u_msg "There was an error whilst trying to determine\nif your player name and password were correct.\n", 0;
        return 127;
	}
}

# == END of start.pl ==
